###################################################################################
#     Estimation of a sparse finite Gaussian mixture model - main file
####################################################################################

## read sources
source("estimation-mixture.R")
source("identification-mixture.R")
source("plot-functions.R")

## load packages
library("e1071")
library("mclust")
library("MASS")
library("bayesm")
library("MCMCpack")
library("mvtnorm")
library("flexclust")

set.seed(123)

##---------- A) read data -----------------------------------------------

## read iris data
data("iris", package = "datasets")
y <- as.matrix(iris[, 1:4])
z <- as.integer(iris$Species) # true classification

###### show data
par(mfrow = c(1, 1))
pairs(y, pch = 20, col = 1)

##---------- B) Specification of the simulation and prior parameters -----------------------------------------------

## number of mixture components
K <- 3
## Dirichlet parameter for the mixture weights
e0 <- 4

## number of iterations, M without burnin
M <- 2000
burnin <- 1000
Mmax <- M + burnin

## read dimensions of data
r <- ncol(y)
N <- nrow(y)

## prior on mu and sigma (Richardson & Green 1997)
## prior on sigma
R <- apply(y, 2, function(x) diff(range(x)))
c0 <- 2.0 + (r - 1)/2
C0 <- 0.75 * cov(y)   # initial value for C0
g0 <- 0.2 + (r - 1)/2
G0 <- 100 * g0/c0 * diag((1/R^2))

## prior on mu
b0 <- apply(y, 2, median)
B0 <- diag(R^2)

## initial values for the parameters:
eta0 <- rep(1/K, K)
sigma0 <- array(0, dim = c(r, r, K))
for (k in 1:K) {
  sigma0[, , k] <- C0
}

## initial classification
groups <- K
cl_y <- kmeans(y, centers = groups, nstart = 30)
S0 <- cl_y$cluster
mu0 <- t(cl_y$centers)   # initial component means

##---------- C) Gibbs sampling from the posterior -----------------------------------------------

## call MCMC procedure
result <- sampling(y, S0, mu0, sigma0, eta0, e0, 
                   c0, C0, g0, G0, b0, B0, M, burnin)

## extract posterior draws
Mu <- result$Mu
Sigma <- result$Sigma
Eta <- result$Eta
S_matrix <- result$S_matrix
Nk_matrix <- result$Nk_matrix
nonnormpost_mode_list <- result$nonnormpost_mode_list

## ---------- D) some diagnostic plots -----------------------------------------------

## Number of nonempty components (K+), after burnin:
Kp <- apply(Nk_matrix, 1, function(x) sum(x != 0))
plot(1:length(Kp), Kp, type = "l", ylim = c(1, K), col="red", lwd=2, 
     xlab = paste("iteration"), main = "K+", ylab="K+")

## Number of observations assigned to component k (Nk):
sort_matrix <- t(apply(Nk_matrix, 1, function(x) sort(x, decreasing = TRUE)))
index <- seq(100, M, by = 100)
Traceplot_Nk(sort_matrix[index, ], label = "Nk")

## point process representation: all mu draws (also from empty components)
pointProcessRepresentation(Mu, label = "mu")

## trace plot of a component mean
k <- 1    # component
j <- 1    # variable
par(mfrow = c(1, 1))
Traceplot_draws(Mu[, j, k, drop = TRUE], label = "mu")

##---------- E) Identification of the mixture model -----------------------------------------------

## estimating K+:
K0_vector <- rowSums(Nk_matrix != 0)  # vector with number of non-empty groups of each iteration
p_K0 <- tabulate(K0_vector, K)/length(K0_vector)
p_K0
par(mfrow = c(1, 1))
barplot(p_K0, names = 1:K, xlab = "K+", col = "lightgrey", ylab = "p(K+|y)")
K0 <- which.max(p_K0)
K0  # mode K0 is the estimator for K+_true
M0 <- sum(K0_vector == K0)
M0  # M0 draws have exactly K0 non-empty groups

## selecting those draws where the number of non-empty groups was exactly K0:
Nk_matrix_K0 <- (Nk_matrix[K0_vector == K0, ] != 0)

## extracting those draws which are sampled from exactly K0 non-empty groups 
## i) Mu:
Mu_inter <- Mu[K0_vector == K0, , ]  # matrix with draws only from the K0 interesting groups
Mu_K0 <- array(0, dim = c(M0, r, K0))  # matrix where empty groups are cut off
for (j in 1:r) {
  Mu_K0[, j, ] <- Mu_inter[, j, ][Nk_matrix_K0]
}

## iii) Eta:
Eta_inter <- Eta[K0_vector == K0, , drop = FALSE] # matrix with draws only from the K0 interesting groups
Eta_K0 <- matrix(Eta_inter[Nk_matrix_K0], ncol = K0)

## iv) S_matrix:
S_matrix_inter <- matrix(0, M, N)
for (m in 1:M) {
  if (K0_vector[m] == K0) {
    perm_S <- rep(0, K)
    perm_S[Nk_matrix[m, ] != 0] <- 1:K0
    S_matrix_inter[m, ] <- perm_S[S_matrix[m, ]]
  }
}
S_matrix_K0 <- S_matrix_inter[K0_vector == K0, ]

## point process representation: mu draws from exactly K+ nonempty components
## before clustering
pointProcessRepresentation(Mu_K0, label = "mu")

## clustering and relabeling of the MCMC draws corresponding to the selected model
map_mu <- nonnormpost_mode_list[[K0]]$bk  # initial values for the group centers
map_cov <- nonnormpost_mode_list[[K0]]$Bk  # initial values for the group covariances
clust_FS_K0 <- identifying(Mu_K0, Eta_K0, S_matrix_K0, map_mu, map_cov, maha = FALSE)
Mu_only_perm <- clust_FS_K0$Mu_only_perm
Eta_only_perm <- clust_FS_K0$Eta_only_perm
S_matrix_only_perm <- clust_FS_K0$S_matrix_only_perm
non_perm_rate <- clust_FS_K0$non_perm_rate
non_perm_rate         # non-permuation rate
Mu_Matrix_FS <- colMeans(Mu_only_perm, dims = 1)  # estimated component means
Eta_Matrix_FS <- colMeans(Eta_only_perm, dims = 1)  # estimated mixture weights

## point process representation: mu draws from exactly Kp nonempty components
## after clustering
pointProcessRepresentation(Mu_only_perm, label = "mu")


##---------- F) Classification of the observations -----------------------------------------------

## assign each observation to the component 
## to which is was assigned mosten often during MCMC sampling
Ass <- matrix(0, N, K0)
for (n in 1:N) {
  Ass[n, ] <- tabulate(S_matrix_only_perm[, n], K0)
}
ass <- apply(Ass, 1, which.max)  # estimed classification
table(ass)                       # estimated cluster sizes


## evaluate the estimated classification
if (!is.null(z)) {
  cat("Classification table:", "\n")
  print(table(z, ass))
  ari <- adjustedRandIndex(z, ass)
  cat("\n", "ARI: ", ari, "\n")
  err <- classError(ass, z)$errorRate
  cat("Error: ", err)
}

##### scatterplot of the estimated classification
pairs(y, pch = 19, col = ass, main = "estimated classification")


