############################################################################
## Gibbs sampling for fitting a finite multivariate Gaussian mixture
############################################################################


sampling <- function(y, S0, mu0, sigma0, eta0, 
                     e0, c0, C0, g0, G0, b0, B0,
                     M, burnin, verbose = FALSE) {
  
  K <- ncol(mu0)  # number of components 
  N <- nrow(y)     # number of observations
  r <- ncol(y)     # number of dimensions
  
  ## initializing current values
  eta0 <- rep(1/K, K)
  sigma0 <- array(0, dim = c(r, r, K))
  mu_j <- matrix(0, r, K)
  S_j <- rep(0, N)
  C0_j <- matrix(0, r, r)
  
  eta_j <- eta0
  sigma_j <- sigma0
  invsigma_j <- sigma_j
  det_invsigma_j <- rep(0, K)
  mu_j <- mu0
  S_j <- S0
  C0_j <- C0
  invB0 <- solve(B0)
  Nk_j <- tabulate(S_j, K)
  if (verbose) {    
      cat("Iteration   0: ", Nk_j)
  }
  
  ## generating matrices for storing the draws
  result <- list(Eta = matrix(0, M, K), 
                 Mu = array(0, dim = c(M, r, K)), 
                 Sigma = array(0, dim = c(M, r, r, K)), 
                 S_matrix = matrix(0L, M, N), 
                 Nk_matrix = matrix(0L, M, K), 
                 Nk_view = matrix(0L, (M + burnin)/100 + 1, K), 
                 mixlik = rep(0, M), 
                 mixprior = rep(0, M),
                 nonnormpost = rep(0, M), 
                 cdpost = rep(0, M), 
                 nonnormpost_mode_list = vector("list", K))
  
  ## initializing the storing matrices
  result$Mu[1, , ] <- mu0
  result$Eta[1, ] <- eta0
  result$S_matrix[1, ] <- S0
  result$Nk_matrix[1, ] <- Nk_j
  for (k in 1:K) {
    result$nonnormpost_mode_list[[k]] <- list(nonnormpost = -(10)^18)
  }
  
  #-------------------------------------MCMC sampling
  
  m <- 2
  while (m <= M | m <= burnin) {
    
    if (m == burnin) {
      m <- 1
      burnin <- 0
    }
    
    ## relevant component specific quantities:
    mean_yk <- matrix(0, r, K)
    mean_yk <- sapply(1:K, function(k) colMeans(y[S_j == k, , drop = FALSE]))
    if (sum(is.na(mean_yk)) > 0) {
      ## to catch the case if a group is empty: NA values are substituted by zeros
      mean_yk[is.na(mean_yk)] <- 0
    }
    if (verbose && !(m%%100)) {
      cat(paste0("\nIteration ", m, ": "), Nk_j)
    }
    K0_j <- sum(Nk_j != 0)  # number of nonempty components
    
    
    #################### first step: parameter simulation (conditional on classification S_j): (1a): Sample eta_j:
    ek <- e0 + Nk_j
    eta_j <- bayesm::rdirichlet(ek)
    
    
    #### (1b): sample Sigma^{-1} for each component k: calculate posterior moments ck and Ck and sample
    #### from the inverted Wishart distribution:
    Ck <- array(0, dim = c(r, r, K))
    ck <- c0 + Nk_j/2
    for (k in 1:K) {
      if (Nk_j[k] != 0) {
        # Bettina:
        Ck[, , k] <- C0_j + 0.5 * crossprod(sweep(y[S_j == k, , drop = FALSE], 2, mu_j[, k], FUN = "-"))
      } else {
        Ck[, , k] <- C0_j
      }
      sig <- rwishart(2 * ck[k], 0.5 * chol2inv(chol(Ck[, , k])))  # attention: rwishart(nu,v)(Rossi)=> nu=2*c0,v=0.5*C0, wishart(c0,C0) (FS)
      sigma_j[, , k] <- sig$IW
      invsigma_j[, , k] <- sig$W
      det_invsigma_j[k] <- det(invsigma_j[, , k])
    }
    
    
    #### (1c): Sample mu_j for each component k:
    Bk <- array(0, dim = c(r, r, K))
    bk <- matrix(0, r, K)
    for (k in 1:K) {
      Bk[, , k] <- chol2inv(chol(invB0 + invsigma_j[, , k] * Nk_j[k]))
      bk[, k] <- Bk[, , k] %*% (invB0 %*% b0 + invsigma_j[, , k] %*% mean_yk[, k] * Nk_j[k])
      mu_j[, k] <- t(chol(Bk[, , k])) %*% rnorm(r) + bk[, k]
    }
    
    
    
    #################### second step: classification of observations (conditional on knowing the parameters) Bettina:
    mat <- sapply(1:K, function(k) eta_j[k] * dmvnorm(y, mu_j[, k], sigma_j[, , k]))
    S_j <- apply(mat, 1, function(x) sample(1:K, 1, prob = x))
    Nk_j <- tabulate(S_j, K)
    
    
    ###################### third step: sample the hyperparameters 
    #### (3a): sample the hyperparameter C0 conditionally on sigma:
    gn <- g0 + K * c0
    C0_j <- rwishart(2 * gn, 0.5 * chol2inv(chol(G0 + rowSums(invsigma_j, dims = 2))))$W  #from package 'bayesm'
    
    
    
    
    
    ###################### fourth step:    
    ## for clustering the draws in point process representation: 
    ## evaluating the mixture likelihood and the complete-data posterior
    if (burnin == 0) {
      ## evaluating the mixture likelihood:
      mat_neu <- sapply(1:K, function(k) eta_j[k] * dmvnorm(y, mu_j[, k], sigma_j[, , k]))
      mixlik_j <- sum(log(rowSums(mat_neu)))
    
      ## evaluating the mixture prior:
      mixprior_j <- log(MCMCpack::ddirichlet(as.vector(eta_j), rep(e0, K))) + 
          sum(dmvnorm(t(mu_j), b0, B0, log = TRUE)) + 
          sum(sapply(1:K, function(k) lndIWishart(2 * c0, 0.5 * C0_j, sigma_j[, , k]))) + 
          lndIWishart(2 * g0, 0.5 * G0, C0_j)  

    
      ## evaluating the nonnormalized complete-data posterior:
      cd <- vector("numeric", N)
      for (k in 1:K) {
        if (sum(S_j == k)) {
            cd[S_j == k] <- dmvnorm(y[S_j == k, ], mu_j[, k], sigma_j[, , k], log = TRUE)
        }
      }
      cdpost_j <- sum(cd) + mixprior_j
      result$mixlik[m] <- mixlik_j
      result$mixprior[m] <- mixprior_j
      result$nonnormpost[m] <- result$mixlik[m] + result$mixprior[m]
      result$cdpost[m] <- cdpost_j
    }
    
    ###################### fourth step: permutation of the labeling and storing the results
    if (burnin == 0) {
      perm <- sample(K)
      ### storing the new values:
      result$Mu[m, , perm] <- mu_j
      result$Eta[m, perm] <- eta_j
      result$S_matrix[m, ] <- perm[S_j]
      result$Nk_matrix[m, perm] <- Nk_j
      result$Sigma[m, , , perm] <- sigma_j
    }
    
    if ((burnin == 0) & (result$nonnormpost[m] > result$nonnormpost_mode_list[[K0_j]]$nonnormpost)) {
        result$nonnormpost_mode_list[[K0_j]] <- list(nonnormpost = result$nonnormpost[m],
                                                     mu = mu_j[, Nk_j != 0],
                                                     bk = bk[, Nk_j != 0], 
                                                     Bk = Bk[, , Nk_j != 0], 
                                                     eta = eta_j[Nk_j != 0])
    }
    
    m <- m + 1
  }
  if (verbose) {
    cat("\n")
  }  
  return(result)
}



dmvnorm <- function(x, mean, sigma, log = FALSE) {
  if (is.vector(x)) {
    x <- matrix(x, ncol = length(x))
  }
  if (missing(mean)) {
    mean <- rep(0, length = ncol(x))
  }
  if (missing(sigma)) {
    sigma <- diag(ncol(x))
  }
  if (NCOL(x) != NCOL(sigma)) {
    stop("x and sigma have non-conforming size")
  }
  if (!isSymmetric(sigma, tol = sqrt(.Machine$double.eps), check.attributes = FALSE)) {
    stop("sigma must be a symmetric matrix")
  }
  if (length(mean) != NROW(sigma)) {
    stop("mean and sigma have non-conforming size")
  }
  distval <- mahalanobis(x, center = mean, cov = chol2inv(chol(sigma)), inverted = TRUE)
  logdet <- sum(log(eigen(sigma, symmetric = TRUE, only.values = TRUE)$values))
  logretval <- -(ncol(x) * log(2 * pi) + logdet + distval)/2
  if (log) 
    return(logretval)
  exp(logretval)
}

